"use strict";

const stars=[];//隕石の場所を保持する配列
let keymap=[];//現在どのキーが押されているかを保持する配列
let ctx;//描画コンテキスト
let ship;//自機のオブジェクト
let score=0;//現在の点数
let speed=25;//スピード
let timer=NaN;//タイマー

//自機のオブジェクトを作るためのコンストラクタ
function Ship(x,y){
    this.x=x;//自機X座標
    this.y=y;//自機Y座標
    //キーの押し下げ状態を保持するメソッド
    this.keydown=(e)=>{
        keymap[e.key]=true;//押されたキーのkeyプロパティをtrueに設定
    };
    this.keyup=(e)=>{
        keymap[e.key]=false;//離したキーのkeyプロパティをfalseに設定
    };
    //キーの押し下げ状態を見て自機を移動させるメソッド
    //左右と上下でif文を分けることにより、斜め方向への移動を可能にしている
    this.move=function(){
        if(keymap["ArrowLeft"]){//左
            this.x-=30;
        }else if(keymap["ArrowRight"]){//右
            this.x+=30;
        }

        if(keymap["ArrowUp"]){//上
            this.y-=30;
        }else if(keymap["ArrowDown"]){//下
            this.y+=30;
        }
        //if文を使わずに１行で変数の範囲を制限する記述方法
        this.x=Math.max(-800,Math.min(800,this.x));//X方向の移動範囲を-800～800に制限
        this.y=Math.max(-800,Math.min(800,this.y));//Y方向の移動範囲を-800～800に制限
    };
}

function random(v){
    return Math.floor(Math.random()*v);//0～v未満の乱数（整数）を返す
}

//ゲームの初期化
function init(){
    //隕石を200個生成し、配列starsに格納
    for(let i=0;i<200;i++){
        //隕石オブジェクトの生成
        stars.push({
            //x,y,zは隕石の初期座標値。rは隕石の初期角度。wは隕石の回転するスピード
            x:random(800*4)-1600,//X座標(-1600～1600)
            y:random(800*4)-1600,//Y座標(-1600～1600)
            z:random(4095),//Z座標(高さ・奥行)(0～4095)
            r:random(360),//角度(0～360°)
            w:random(10)-5//回転スピード(-5～5)
        });
    }

    ship=new Ship(200,200);//自機作成
    onkeydown=ship.keydown;//キー押下時のイベントハンドラ登録
    onkeyup=ship.keyup;//キーを離した時のイベントハンドラ登録

    let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
    let space = document.getElementById("space");//<canvas>要素への参照を取得
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    space.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    space.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = space.getContext("2d");//描画コンテキスト
    ctx.font = "3rem Arial";
    repaint();
}

//STARTボタンを押した時に呼び出され、実際にゲームを開始する関数
function go(){
    let space = document.getElementById("space");
    space.onmousedown=mymousedown;//マウス押下時イベントハンドラ登録
    space.onmouseup=mymouseup;//マウスリリース時イベントハンドラ登録
    space.oncontextmenu=(e)=>{
        e.preventDefault();//コンテキストメニュー非表示に
    };
    space.addEventListener("touchstart",mymousedown);
    space.addEventListener("touchend",mymouseup);

    //タッチで指を動かした時、ブラウザによっては、ゲーム操作ではなく、ゲーム画面を移動する操作と解釈するため記述
    document.body.addEventListener(
        "touchmove",
        (event)=>{
            event.preventDefault();//画面タッチによるコンテキストメニューを非表示に
        },
        false
    );
    document.getElementById("START").style.display="none";//STARTボタンを非表示に
    document.getElementById("bgm").play();//BGM再生開始
    timer=setInterval(tick,50);
}

function mymousedown(e){
    //マウスとタッチのイベント処理 --> 座標に応じて上下左右キーのkeymapを更新
    let mouseX=(!isNaN(e.offsetX)?e.offsetX:e.touches[0].clientX)-space.width/2;
    let mouseY=(!isNaN(e.offsetY)?e.offsetY:e.touches[0].clientY)-space.height/2;
    if(Math.abs(mouseX)>Math.abs(mouseY)){
        keymap[mouseX>0?"ArrowLeft":"ArrowRight"]=true;
    }else{
        keymap[mouseY>0?"ArrowUp":"ArrowDown"]=true;
    }
}

function mymouseup(e){
    keymap=[];
}

//メインループ
function tick(){
    stars.forEach((s) => {
        s.z-=speed;//隕石のZ方向の値を変数speedの値だけ減らす(自機の方へ隕石を近づける)
        s.r+=s.w;//隕石を回転させる
        //衝突判定
        if(s.z<64){//隕石がほぼ自機と同じ平面に到達したかを判定(自機はz=0の平面上を移動)
            if(Math.abs(s.x-ship.x)<50 && Math.abs(s.y-ship.y)<50){//隕石と自機のX,Y座標の差が共に50未満の時
                //衝突-->ゲームオーバー
                clearInterval(timer);//タイマーを止める
                timer=NaN;
                document.getElementById("bgm").pause();//BGMを止める
                return;
            }
            //衝突しなかった場合は、自機が動く平面を通過-->一番奥へ隕石を再配置
            s.x=random(800*4)-1600;
            s.y=random(800*4)-1600;
            s.z=4095;
        }
    });

    if(score++%10==0){//変数scoreを増やして、その値が10の倍数の場合
        speed++;
    }
    ship.move();//自機を移動
    repaint();//画面を再描画
}

function repaint(){
    space.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    space.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる

    ctx.fillStyle="black";
    ctx.fillRect(0,0,space.width,space.height);//背景クリア
    //sort()メソッド:関数オブジェクトの結果に基づいて配列の要素の並べ替えを行う。引数に比較関数を指定しない場合には文字列として比較される
    //引数に比較関数を指定する場合、引数を２個受け取る必要がある
    //引数が(a,b)の場合に、ａ⇒ｂの順番に並べる場合は０より小さい戻り値を返し、ｂ⇒ａの順番に並べる場合は０より大きい戻り値を返す
    stars.sort((a,b)=>b.z-a.z)//隕石を距離順(Z軸の値順)に並べ替え＝奥の隕石から描画される様に並べ替え

    //隕石の描画
    stars.forEach((s)=>{
        let z=s.z;
        //広がり(物体の遠近)を生み出すため、隕石と自機の差分をZ座標(高さ・奥行)で割り、遠くの隕石ほど画面中央からのX・Y座標の差分を小さくする
        let x=((s.x-ship.x)*512)/z+space.width/2;//「+space.width/2」は、描画領域の中心を原点とするためのもの
        let y=((s.y-ship.y)*512)/z+space.height/2;//「+space.height/2」は、描画領域の中心を原点とするためのもの
        let size=(50*512)/z;//距離に応じてサイズを設定
        ctx.save();//現在のコンテキストを保存
        ctx.translate(x,y);//(x,y)へ座標系を並行移動
        //globalAlphaプロパティ：透明度(opacity)を設定するプロパティで、値が１に近いほど透過されず元画像に近い状態になり、値が０に近いほど透過され透明に描画される
        ctx.globalAlpha=1-z/4096;//遠くの隕石ほど暗く描画する(ほとんど透明に描画して背景色の黒に溶け込ませる手法)
        ctx.rotate((s.r*Math.PI)/180);//座標系を回転
        ctx.drawImage(rockImg,-size/2,-size/2,size,size);
        ctx.restore();//保存したコンテキストを復元
    });

    ctx.drawImage(scope,space.width/4,0,space.width/2,space.height);//コックピット画像
    ctx.fillStyle="purple";
    ctx.font="3rem Arial";
    //スコアの左を０で埋めて描画するための処理
    //slice()メソッドの引数にマイナスの値を指定した場合、文字列の末尾から切り取って返される
    ctx.fillText(("0000000"+score).slice(-7),space.width/2.3,space.height/10);
    if(isNaN(timer)){
        ctx.fillStyle="#5aFFFF";
        ctx.fillText("GAME OVER",space.width/2.5,space.height/2);
    }
}