"use strict";

//DotクラスとEatクラスの継承元
class Sprite {
    constructor(x, y, r) {
        this.x = x;//X座標
        this.y = y;//Y座標
        this.sx = 0;//X方向速度
        this.sy = 0;//Y方向速度
        this.r = r;//半径
        this.count = 0;//カウンタ
    }
    /*cos()関数は、移動方向のθ(シータ、移動方向の角度)からX軸方向成分を求める関数
      sin()関数は、移動方向のθ(シータ、移動方向の角度)からY軸方向成分を求める関数
      atan2()関数は、X軸方向成分とY軸方向成分から移動方向のθ(シータ、移動方向の角度)を求める関数*/
    //countの値を徐々に増やしながら、その値に応じて口をパクパク開け閉めするアニメーションを描画している
    draw() {
        this.count += 0.5;
        let t = Math.sin(this.count) * 0.5;
        let d = Math.atan2(this.sy, this.sx);//移動方向の角度(θ)
        ctx.beginPath();
        ctx.moveTo(this.x, this.y);
        ctx.arc(this.x, this.y, this.r, d + t, d + Math.PI * 2 - t);
        ctx.closePath();
        ctx.fill();
    }
}

//追いかける側のクラス
class Eat extends Sprite {
    //tick()メソッドは、アニメーションのフレーム(画面)を描画するたびに呼び出される
    tick() {
        this.sx += (mouse.x - this.x) / 50;//マウスの座標と現在の座標の差分に応じて速度を加減
        this.sy += (mouse.y - this.y) / 50;
        this.sx *= 0.98;//徐々に減速
        this.sy *= 0.98;
        this.x += this.sx;//速度を座標に反映
        this.y += this.sy;
        this.draw();
    }
}

//追われる側のクラス
class Dot extends Sprite {
    constructor() {
        //初期座標。super()関数は、親クラスのコンストラクタを呼び出す
        super(Math.random() * (field.width - 100) + 50, Math.random() * (field.height - 100) + 50, 10);
        //初期速度
        this.sx = Math.random() * 10 - 5;
        this.sy = Math.random() * 10 - 5;
    }

    //tick()メソッドは、アニメーションのフレーム(画面)を描画するたびに呼び出される
    tick() {
        this.x = (this.x + this.sx + field.width) % field.width;//X座標を更新
        this.y = (this.y + this.sy + field.height) % field.height;//Y座標を更新
        this.draw();
    }
}

let ctx;//描画コンテキスト
let dots = [];//餌の配列
let life = 600;//残り時間
let timerId = NaN;//タイマー
let back;//背景画像
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let field = document.getElementById("field");//<canvas>要素への参照を取得
const mouse = { x: 0, y: 0 };//マウスの座標
const eat = new Eat(300, 300, 30);//自分(追いかける側)のオブジェクト

//初期化
onload = () => {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    field.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    field.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = field.getContext("2d");//描画コンテキスト
    ctx.font = "2rem 'Times New Roman'";
    //マウス移動イベントハンドラの登録
    window.onmousemove = (e) => {
        mouse.x = e.clientX;
        mouse.y = e.clientY;
    };
    for (let i = 0; i < 15; i++) {
        dots.push(new Dot());//15個の餌(Dot)を作成して、配列dotsに格納
    }
    back = document.getElementById("back");
    timerId = setInterval(tick, 50);//タイマー開始
};

//メインルーチン
function tick() {
    ctx.drawImage(back, 0, 0, field.width, field.height);//背景画像を描画
    eat.tick();//自分を更新 移動と描画

    dots.forEach((d) => {
        ctx.fillStyle = "#EA0032";
        d.tick();//餌(dotオブジェクト)を更新 移動と描画
    });

    ctx.fillStyle = "#aa0";

    //衝突判定：自分とのxとyの距離の差が30以上の餌のみfilter()メソッドでフィルタリングして取り出す
    dots = dots.filter((d) => {
        return Math.abs(eat.x - d.x) > 30 || Math.abs(eat.y - d.y) > 30
    });

    life -= 2;
    ctx.fillRect(0, 0, life, 5);
    ctx.fillText(life + 2, field.width / 2, field.height / 20);
    //ゲームオーバー
    if (life < 0) {
        clearInterval(timerId);
        ctx.fillText("ゲームオーバー", field.width / 2.3, field.height / 2);
    }
    //ゲームクリア
    if (dots.length == 0) {
        clearInterval(timerId);
        ctx.fillText("ゲームクリア！", field.width / 2.3, field.height / 2);
    }
}
