"use strict";

let ctx;//描画コンテキスト
let engine;//物理エンジンオブジェクト
let target;//選択中の玉
let mousePos = null;//マウス座標
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得
canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる

let walls = [
    [-100 + canvas.width / 4.5, -100 + canvas.height / 18, 1000 + canvas.width / 4.5, 140 + canvas.height / 18],
    [-100 + canvas.width / 4.5, 410 + canvas.height / 9, 1000 + canvas.width / 4.5, 100 + canvas.height / 9],
    [-100 + canvas.width / 1000, -100 + canvas.height / 18, 140 + canvas.width / 4.5, 650 + canvas.height / 18],
    [760 + canvas.width / 4.5, -100 + canvas.height / 18, 100 + canvas.width / 4.5, 650 + canvas.height / 18],
];//上下左右の壁の座標情報

let holes = [
    [35 + canvas.width / 4.5, 35 + canvas.height / 9],
    [400 + canvas.width / 4.5, 35 + canvas.height / 9],
    [765 + canvas.width / 4.5, 35 + canvas.height / 9],
    [35 + canvas.width / 4.5, 415 + canvas.height / 9],
    [400 + canvas.width / 4.5, 415 + canvas.height / 9],
    [765 + canvas.width / 4.5, 415 + canvas.height / 9],
];//6つの穴(ポケット)の座標情報

let balls = [
    { x: 200 + canvas.width / 4.5, y: 200 + canvas.height / 9, c: "#FFF400" },
    { x: 125 + canvas.width / 4.5, y: 185 + canvas.height / 9, c: "#005CD3" },
    { x: 150 + canvas.width / 4.5, y: 170 + canvas.height / 9, c: "#CE2721" },
    { x: 100 + canvas.width / 4.5, y: 200 + canvas.height / 9, c: "#BD4CB8" },
    { x: 175 + canvas.width / 4.5, y: 215 + canvas.height / 9, c: "#F06700" },
    { x: 125 + canvas.width / 4.5, y: 215 + canvas.height / 9, c: "#0B8A17" },
    { x: 175 + canvas.width / 4.5, y: 185 + canvas.height / 9, c: "#B70D3A" },
    { x: 150 + canvas.width / 4.5, y: 230 + canvas.height / 9, c: "#333333" },
    { x: 150 + canvas.width / 4.5, y: 200 + canvas.height / 9, c: "#FFD300" },
    { x: 650 + canvas.width / 4.5, y: 200 + canvas.height / 9, c: "#CAFDFF" },
];//玉の座標情報

onload = function init() {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    //エンジン初期化
    engine = new Engine(-100, -100, canvas.width, canvas.height, 0, 0);
    //イベントハンドラ登録
    canvas.onmousedown = mymousedown;
    canvas.onmousemove = mymousemove;
    canvas.onmouseup = mymouseup;
    canvas.addEventListener("touchstart", mymousedown);
    canvas.addEventListener("touchmove", mymousemove);
    canvas.addEventListener("touchend", mymouseup);

    //壁を作成してエンジンに追加
    walls.forEach((w) => {
        let r = new RectangleEntity(w[0], w[1], w[2], w[3]);
        r.color = "gray";
        engine.entities.push(r);
    });

    //玉を作成してエンジンに追加
    balls.forEach((b) => {
        let r = new CircleEntity(b.x, b.y, 20, BodyDynamic, 0.9, 0.99);
        r.color = b.c;
        //配列の個々の要素(b)のentityプロパティに物理世界の玉オブジェクト(r)を格納する
        b.entity = r;
        engine.entities.push(r);
    });

    //穴を作成してエンジンに追加
    holes.forEach((h) => {
        let r = new CircleEntity(h[0], h[1], 20, BodyStatic);//穴のオブジェクト
        r.color = "rgba(255,255,255,0)";
        //穴のオブジェクトにonhit()メソッド(自身が他のオブジェクトと衝突した時に呼び出される)を追加
        r.onhit = (me, peer) => {
            //穴(me)に衝突した玉(peer)を削除
            engine.entities = engine.entities.filter((e) => {
                return e != peer;//衝突相手以外の要素を抽出(trueを返す)
            });
        };
        engine.entities.push(r);
    });

    //その他(Canvas, Timer)の初期化
    ctx = canvas.getContext("2d");
    ctx.font = "2rem Arial";
    ctx.strokeStyle = "blue";
    setInterval(tick, 50);
};

function tick() {
    engine.step(0.01);//物理エンジンの時刻を進める
    repaint();//再描画
}

//マウス押し下げ時のコールバック関数
function mymousedown(e) {
    let mouseX = !isNaN(e.offsetX) ? e.offsetX : e.touches[0].clientX;
    let mouseY = !isNaN(e.offsetY) ? e.offsetY : e.touches[0].clientY;
    for (let i = 0; i < balls.length; i++) {
        //マウス押し下げ時の座標が玉に含まれている場合
        if (balls[i].entity.isHit(mouseX, mouseY)) {
            //マウス押し下げ時の座標を含む玉をtarget(手玉)に設定
            target = balls[i].entity;
            mousePos = { x: mouseX, y: mouseY };//マウス押し下げ時の座標
            break;
        }
    }
}

//マウス移動時のコールバック関数
function mymousemove(e) {
    let mouseX = !isNaN(e.offsetX) ? e.offsetX : e.touches[0].clientX;
    let mouseY = !isNaN(e.offsetY) ? e.offsetY : e.touches[0].clientY;
    if (target) {
        mousePos = { x: mouseX, y: mouseY };//マウスの座標を更新
    }
}

//マウスリリース時のコールバック関数
function mymouseup(e) {
    if (target) {
        //現在のマウス座標と手玉の座標の差分
        let dx = mousePos.x - target.x;
        let dy = mousePos.y - target.y;
        //手玉にX方向・Y方向の初速度を設定
        target.velocity.x = dx / 10;
        target.velocity.y = dy / 10;
    }
    target = null;
}

function repaint() {
    ctx.drawImage(billiard, canvas.width / 4.5, canvas.height / 9, 800, 450);//背景クリア
    //マウスをドラッグした時に表示される線をマウスリリース後に見えなくする
    ctx.clearRect(0, 0, canvas.width, canvas.height / 9);//上
    ctx.clearRect(0, 450 + canvas.height / 9, canvas.width, canvas.height);//下
    ctx.clearRect(0, 0, canvas.width / 4.5, canvas.height);//左
    ctx.clearRect(800 + canvas.width / 4.5, 0, canvas.width, canvas.height);//右

    //ボール・壁の描画
    for (let i = 0; i < engine.entities.length; i++) {
        let e = engine.entities[i];
        ctx.fillStyle = e.color;
        switch (e.shape) {
            case ShapeCircle://円の場合
                ctx.beginPath();
                ctx.arc(e.x, e.y, e.radius, 0, Math.PI * 2);
                ctx.closePath();
                ctx.fill();
                break;
        }
    }

    if (target && mousePos) {
        ctx.beginPath();
        ctx.moveTo(target.x, target.y);
        ctx.lineTo(mousePos.x, mousePos.y);
        ctx.stroke();
    }
}
