"use strict";

let engine;//物理エンジンオブジェクト
let ctx;//描画コンテキスト
let colors = ["yellow", "green", "orange", "blue", "white"];//色の配列
let fruits = [];//画像の配列
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得

function rand(v) {
    return Math.floor(Math.random() * v);//0～vまでの乱数(整数)を返す
}

//初期化関数
function init() {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    fruits = document.querySelectorAll("img");
    let r;
    engine = new Engine(0, 0, canvas.width, canvas.height, 0, 9.8);//物理エンジンオブジェクト作成

    //矩形を作成しエンジンに追加
    r = new RectangleEntity((canvas.width) - 50, 50, 50, canvas.height);
    r.color = "green";
    engine.entities.push(r);//作成したオブジェクトを物理世界に追加

    r = new RectangleEntity(0, 50, 50, canvas.height);
    r.color = "yellow";
    engine.entities.push(r);

    //線分を作成しエンジンに追加
    r = new LineEntity(50, 300, (canvas.width) - 50, canvas.height / 1.2);
    r.color = "orange";
    engine.entities.push(r);

    r = new LineEntity((canvas.width) - 100, canvas.height / 4, 100, 450);
    r.color = "orange";
    engine.entities.push(r);

    //7 x 3 = 21 個の円(固定)を作成
    for (let i = 0; i < 7; i++) {
        for (let j = 0; j < 3; j++) {
            r = new CircleEntity(i * canvas.width / 7 + 100, j * canvas.height / 3 + 100, 5, BodyStatic);
            r.color = colors[j];
            engine.entities.push(r);
        }
    }

    //20 個の円(移動)を作成
    for (let i = 0; i < 20; i++) {
        r = new CircleEntity(rand(canvas.width - 100) + 50, rand(canvas.height / 2), 10, BodyDynamic);
        r.color = colors[rand(5)];
        r.fruit = fruits[rand(5)];
        r.velocity.x = rand(10) - 5;//円オブジェクトのX軸方向初速度
        r.velocity.y = rand(10) - 5;//円オブジェクトのY軸方向初速度
        engine.entities.push(r);
    }

    ctx = canvas.getContext("2d");//描画コンテキスト
    setInterval(tick, 50);
}

//メインループ
function tick() {
    engine.step(0.01);//物理エンジンの時刻を進める
    repaint();
}

function repaint() {
    ctx.fillStyle = "black";
    ctx.fillRect(0, 0, canvas.width, canvas.height);
    engine.entities.forEach((e) => {
        ctx.fillStyle = e.color;
        ctx.strokeStyle = e.color;
        switch (e.shape) {
            case ShapeRectangle://矩形の場合
                ctx.fillRect(e.x, e.y, e.w, e.h);
                break;
            case ShapeCircle://円の場合
                if (e.fruit) {
                    ctx.drawImage(e.fruit, e.x - 10, e.y - 10, 30, 30)
                } else {
                    ctx.beginPath();
                    ctx.arc(e.x, e.y, e.radius, 0, Math.PI * 2);
                    ctx.closePath();
                    ctx.fill();
                }
                break;
            case ShapeLine://線分の場合
                ctx.beginPath();
                ctx.moveTo(e.x0, e.y0);
                ctx.lineTo(e.x1, e.y1);
                ctx.stroke();
                break;
        }
    });
}
