"use strict";

let ctx;//描画コンテキスト
let engine;//物理エンジン
let veges = [];//選択された野菜オブジェクトを格納する配列
let timer = NaN;//タイマー
let startTime = NaN;//ゲーム開始時刻
let elapsed = 0;//経過時間
let score = 0;//スコア
let walls = [
    [-60, -100, 100, 800],
    [500, -100, 100, 800],
    [-60, 520, 700, 100],
];//壁(左右底)オブジェクトのx, y, w, h座標
let images = [];//野菜の画像を格納する配列
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得

function rand(v) {
    return Math.floor(Math.random() * v);
}

//エンジン初期化 & Canvas初期化
function init() {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = canvas.getContext("2d");//Canvasのコンテキスト設定
    ctx.font = "2rem Arial";
    ctx.strokeStyle = "blue";
    ctx.lineWidth = 5;
    ctx.textAlign = "center";

    //物理エンジンオブジェクトの作成
    engine = new Engine(-100, -100, 700, 700, 0, 9.8);

    //壁オブジェクトの作成
    walls.forEach((w) => {
        let r = new RectangleEntity(w[0], w[1], w[2], w[3]);
        r.color = "gray";
        engine.entities.push(r);
    });

    //野菜オブジェクトの作成
    for (let i = 0; i < 7; i++) {
        for (let j = 0; j < 10; j++) {
            let x = i * 60 + 75 + rand(5);
            let y = j * 50 + 50 + rand(5);
            let r = new CircleEntity(x, y, 25, BodyDynamic, 1, 0.98);
            r.color = rand(5);
            engine.entities.push(r);
        }
    }

    for (let i = 0; i < 5; i++) {
        //野菜画像を配列に格納
        images.push(document.getElementById("fruit" + i));
    }

    repaint();
}

//ゲーム開始時に呼び出される関数(STARTボタンのonclick属性から呼び出される)
function go() {
    //キャンバスにイベントハンドラ登録
    canvas.onmousedown = mymousedown;
    canvas.onmousemove = mymousemove;
    canvas.onmouseup = mymouseup;
    canvas.addEventListener("touchstart", mymousedown);
    canvas.addEventListener("touchmove", mymousemove);
    canvas.addEventListener("touchend", mymouseup);

    //タッチ時のコンテキストメニュー非表示に
    document.body.addEventListener("touchmove", (e) => {
        e.preventDefault();
    }, false);
    //STARTボタンを非表示に
    document.getElementById("START").style.display = "none";
    document.getElementById("bgm").play();//BGMの再生を開始

    startTime = new Date();
    timer = setInterval(tick, 50);
}

//メインループ
function tick() {
    engine.step(0.01);//物理エンジンの時刻を進める

    //経過秒数を求める
    elapsed = (new Date().getTime() - startTime) / 1000;
    //57秒(BGMの再生時間)を過ぎたらタイマーを停止
    if (elapsed > 57) {
        clearInterval(timer);
        timer = NaN;
    }
    repaint();//再描画
}

//マウス押下時の処理
function mymousedown(evt) {
    //マウス押下時のX座標
    let x = !isNaN(evt.offsetX) ? evt.offsetX : evt.touches[0].clientX;
    //マウス押下時のY座標
    let y = !isNaN(evt.offsetY) ? evt.offsetY : evt.touches[0].clientY;
    //マウス押下時の座標で円が選択されたら、野菜とみなして配列vegesに追加
    engine.entities.forEach((e) => {
        if (e.isHit(x, y) && e.shape == ShapeCircle) {
            veges.push(e);
            e.selected = true;
        }
    });
}

//マウス移動時の処理
function mymousemove(evt) {
    if (veges.length == 0) {
        return;
    }

    //マウス移動時のX座標
    let x = !isNaN(evt.offsetX) ? evt.offsetX : evt.touches[0].clientX;
    //マウス移動時のY座標
    let y = !isNaN(evt.offsetY) ? evt.offsetY : evt.touches[0].clientY;
    let p = veges[veges.length - 1];//配列vegesにある最後の野菜を取得

    engine.entities.forEach((e) => {
        //対象となるオブジェクトにマウスの座標が含まれており、かつ、形状が円か否か判断している
        if (e.isHit(x, y) && e.shape == ShapeCircle) {
            //既に野菜が選択されていないか(配列vegesに含まれていないか＝veges.indexOf(e) < 0)、かつ、最後の野菜と今の野菜が同じ種類か(e.color == p.color)を判断している
            if (veges.indexOf(e) < 0 && e.color == p.color) {
                let d2 = Math.pow(e.x - p.x, 2) + Math.pow(e.y - p.y, 2);//最後の野菜との距離
                //最後の野菜と今の野菜の距離が4000未満であれば連結→配列vegesに追加して、野菜を選択状態に
                if (d2 < 4000) {
                    veges.push(e);
                    e.selected = true;
                }
            }
        }
    });
}

//マウスリリース時の処理
function mymouseup(evt) {
    //選択状態の野菜が２つ以上ある場合
    if (veges.length > 1) {
        //非選択状態の野菜だけを抽出する(＝選択状態の野菜を物理世界から削除)
        engine.entities = engine.entities.filter((e) => {
            return e.selected != true;
        });

        //新たな野菜オブジェクトを消去した個数分追加
        for (let i = 0; i < veges.length; i++) {
            let x = 75 + rand(350);
            let r = new CircleEntity(x, 0, 25, BodyDynamic, 1, 0.98);
            r.color = rand(5);
            engine.entities.push(r);
        }
        //選択された野菜オブジェクトの数に応じてスコアを加算
        score += veges.length * 100;
    }
    //全ての野菜オブジェクトのselectedプロパティを削除
    veges.forEach((e) => {
        delete e.selected;
    });
    veges = [];
}

function repaint() {
    ctx.drawImage(fruitbg, 0, 0);//背景をクリア

    //野菜を描画
    for (let i = 0; i < engine.entities.length; i++) {
        let e = engine.entities[i];
        let img = images[e.color];
        if (e.shape == ShapeCircle) {
            ctx.drawImage(img, e.x - 28, e.y - 28, 62, 62);
            //選択状態にある場合(e.selected==true)、円を強調描画して選択されていることが視覚的に分かりやすくする
            if (e.selected) {
                ctx.strokeStyle = "yellow";
                ctx.beginPath();
                ctx.arc(e.x, e.y, e.radius, 0, Math.PI * 2);
                ctx.closePath();
                ctx.stroke();
            }
        }
    }

    //選択状態にある円の中心を線で結んで描画する
    if (veges.length > 0) {
        ctx.strokeStyle = "#B1EB22";
        ctx.beginPath();
        ctx.moveTo(veges[0].x, veges[0].y);
        for (let i = 0; i < veges.length; i++) {
            ctx.lineTo(veges[i].x, veges[i].y);
        }
        ctx.stroke();
    }

    //メッセージ
    ctx.save();
    ctx.fillStyle = "#F9D79F";
    ctx.font = "bold 2rem sans-serif";
    ctx.font;
    ctx.translate(650, 442);
    ctx.rotate(-0.05);
    ctx.fillText(isNaN(timer) ? "FINISH" : "SCORE", 0, 0);
    ctx.restore();

    //スコア
    ctx.save();
    ctx.font = "bold 3rem sans-serif";
    ctx.translate(660, 365);
    ctx.rotate(0.08);
    ctx.fillStyle = "#F9D79F";
    ctx.fillText(("0000000" + score).slice(-7), 0, 0);
    ctx.restore();

    //残り時間
    ctx.save();
    ctx.fillStyle = "rgba(215, 130, 40, 0.5)";
    ctx.beginPath();
    ctx.moveTo(656, 153);
    ctx.arc(
        656,
        153,
        88,
        -Math.PI / 2,
        (elapsed / 57) * Math.PI * 2 - Math.PI / 2
    );
    ctx.closePath();
    ctx.fill();
    ctx.restore();
    ctx.fillText(Math.floor(elapsed) + "秒", 650, 150);
}
